dnl Macros for configuring WHIZARD subdirectories and subpackages

dnl Define a variable and export it
dnl AC_WHIZARD_SET(variable, value)
AC_DEFUN(AC_WHIZARD_SET,
[$1=$2
AC_SUBST($1)])

dnl Add a list of names to the list of output resp. executable files
dnl AC_WHIZARD_OUTFILES(subdir, files)
AC_DEFUN(AC_WHIZARD_OUTFILES,
[for file in $2; do
  OUTFILES="$OUTFILES $1/$file";
done])

dnl AC_WHIZARD_EXECUTABLES(subdir, files)
AC_DEFUN(AC_WHIZARD_EXECUTABLES,
[for file in $2; do
  OUTFILES="$OUTFILES $1/$file";
  EXECUTABLES="$EXECUTABLES $1/$file"
done])


dnl This adds a package which resides in a WHIZARD subdirectory.
dnl The `variable' is inserted into AC_SUBST; it will refer to 
enl the package subdirectory in Makefiles.  The package
dnl has a subdirectory XXX-src and optionally a Makefile (or similar).
dnl AC_WHIZARD_PACKAGE(variable, identifier [,Makefiles [,Executables]])
AC_DEFUN(AC_WHIZARD_PACKAGE,
[AC_WHIZARD_SET($1,$2-src)
ifelse($#, 3, 
[AC_WHIZARD_OUTFILES($$1, $3)])
ifelse($#, 4, 
[AC_WHIZARD_OUTFILES($$1, $3)
AC_WHIZARD_EXECUTABLES($$1, $4)])
])

dnl This is similar, but no '-src' is appended for the directory name
dnl AC_WHIZARD_SUBDIR(variable, identifier [,Makefiles [,Executables]])
AC_DEFUN(AC_WHIZARD_SUBDIR,
[AC_WHIZARD_SET($1,$2)
ifelse($#, 3,
[AC_WHIZARD_OUTFILES($$1, $3)])
ifelse($#, 4, 
[AC_WHIZARD_OUTFILES($$1, $3)
AC_WHIZARD_EXECUTABLES($$1, $4)])
])


dnl This adds a matrix element generating package.
dnl There must be at least one Makefile.
dnl In addition to the preceding macro, it adds two scripts
dnl init.XXX and make-proc.XXX where XXX is the package identifier.
dnl AC_WHIZARD_ME_GENERATOR(variable, identifier, Makefile)
AC_DEFUN(AC_WHIZARD_ME_GENERATOR,
[AC_WHIZARD_PACKAGE($1,$2,$3, init.$2 make-proc.$2)])


dnl This is like AC_WHIZARD_PACKAGE, but it calls AC_ARG_ENABLE in addition.
dnl If the package `id' is disabled or the 'file' is not found,
dnl the variable `var' is set to "no".
dnl AC_WHIZARD_PACKAGE_ENABLE(var, id, help, file [,Makefiles [,Executables]])
AC_DEFUN(AC_WHIZARD_PACKAGE_ENABLE,
[AC_ARG_ENABLE($2,[$3])
if test "$enable_$2" = "no"; then
AC_MSG_CHECKING([for $2-src/$4])
AC_WHIZARD_SET($1, no)
AC_MSG_RESULT([(disabled)])
else
AC_CHECK_PROG(enable_$2, $2-src/$4, yes, no, .)
if test "$enable_$2" = "no"; then
AC_WHIZARD_SET($1, no)
else
ifelse($#, 4, [AC_WHIZARD_PACKAGE($1, $2)])
ifelse($#, 5, [AC_WHIZARD_PACKAGE($1, $2, $5)])
ifelse($#, 6, [AC_WHIZARD_PACKAGE($1, $2, $5, $6)])
fi
fi
])

dnl The same for a matrix element generator
dnl There must be at least one Makefile.
dnl AC_WHIZARD_ME_GENERATOR_ENABLE(var, id, help, file, Makefiles [,Executables])
AC_DEFUN(AC_WHIZARD_ME_GENERATOR_ENABLE,
[AC_WHIZARD_PACKAGE_ENABLE($1, $2, [$3], $4, $5, init.$2 make-proc.$2 $5)])



dnl Extension of AC_PATH_PROG: Search for libraries for which the name
dnl is not exactly known (because it may have the version number in it)
dnl Set output variables $var, $var_DIR, $var_LIB accordingly
dnl WHIZARD_PATH_LIB(var, id, name-list, path)
AC_DEFUN(WHIZARD_PATH_LIB,
[AC_CACHE_CHECK(for $3, ac_cv_path_$1,
[case "$$1" in
  /*)
  ac_cv_path_$1="$$1" # User-supplied path
  ;;
  *)
  IFS="${IFS= 	}"; ac_save_ifs="$IFS"; IFS=":"
  ac_dummy=$4
  for ac_dir in $ac_dummy; do
    test -z "$ac_dir" && ac_dir=.
    unset ac_cv_path_$1
    ac_pwd=`pwd`
    if test -d "$ac_dir"; then
      cd $ac_dir
      for ac_word in $3; do
        test -f "$ac_word" && ac_cv_path_$1="$ac_dir/$ac_word"
      done
      cd $ac_pwd
    fi
    if test -n "$ac_cv_path_$1"; then
      break
    fi
  done
  IFS="$ac_save_ifs"
  if test -z "$ac_cv_path_$1"; then
    ac_cv_path_$1="no"
  fi
  ;;
esac
])
$1=$ac_cv_path_$1
if test "$$1" != "no"; then
$1_DIR=`echo $$1 | sed -e 's|/lib$2.*\.a$||'`
$1_LIB=`echo $$1 | sed -e 's|^.*/lib\($2.*\)\.a$|\1|'`
fi
AC_SUBST($1)
AC_SUBST($1_DIR)
AC_SUBST($1_LIB)
])


dnl Library search with enable option
dnl WHIZARD_LIB_ENABLE(var, id, help, name-list, path)
AC_DEFUN(WHIZARD_LIB_ENABLE,
[AC_ARG_ENABLE($2,[$3])
if test "$enable_$2" = "no"; then
AC_MSG_CHECKING(for $4)
$1=no
AC_MSG_RESULT([(disabled)])
else
WHIZARD_PATH_LIB($1, $2, $4, $5)
fi
])


dnl ------------------------------------------------------------------------
dnl Compiler Tests

dnl These are slightly modified version of the default test for f77

dnl Determine a Fortran 90 compiler to use.  If `F90' is not already set,
dnl check for `f95', `f90', `ifc' and `pgf90', in that order.
dnl Set the output variable `F90' to the name of the compiler found.
dnl 
dnl AC_PROG_F90()
AC_DEFUN(AC_PROG_F90,
[AC_BEFORE([$0], [AC_PROG_CPP])dnl
if test -z "$F90"; then
  AC_PATH_PROGS(F90, f95 f90 ifc pgf90)
  test -z "$F90" && AC_MSG_ERROR([no acceptable Fortran 90 compiler found in \$PATH])
fi
AC_PROG_F90_WORKS
])

dnl Test whether the Fortran 90 compiler can compile and link a trivial
dnl Fortran program.  
dnl Disabled: Also, test whether the Fortran 90 compiler is a cross-compiler.
dnl 
dnl AC_PROG_F90_WORKS()
AC_DEFUN(AC_PROG_F90_WORKS,
[AC_CACHE_CHECK([whether the F90 compiler ($F90 $F90FLAGS $LDFLAGS) works],
[ac_cv_prog_f90_works],
[AC_LANG_SAVE
AC_LANG_FORTRAN90
AC_TRY_COMPILER_NOINCLUDE(dnl
[program conftest
end program conftest
], ac_cv_prog_f90_works, ac_cv_prog_f90_cross)
AC_LANG_RESTORE
])

if test $ac_cv_prog_f90_works = no; then
  AC_MSG_ERROR([installation or configuration problem: Fortran 90 compiler cannot create executables.])
fi

dnl This value is either in the cache or has been determined above.
dnl AC_CACHE_CHECK([whether this compiler is a cross-compiler],
dnl [ac_cv_prog_f90_cross],
dnl [:])

AC_CACHE_CHECK([the compiler vendor],
[ac_cv_prog_f90_type],
[$F90 -version >conftest.log 2>&1
$F90 -V >>conftest.log 2>&1

ac_grep_f90_NAG=`grep NAGWare conftest.log | head -1`
ac_grep_f90_Compaq=`grep Compaq conftest.log | head -1`
ac_grep_f90_Digital=`grep DIGITAL conftest.log | head -1`
ac_grep_f90_SGI=`grep MIPS conftest.log | head -1`
ac_grep_f90_Intel=`grep 'Intel(R)' conftest.log | head -1`

if test -n "$ac_grep_f90_NAG"; then
  ac_cv_prog_f90_type="NAG"
  ac_cv_prog_f90_version=$ac_grep_f90_NAG
elif test -n "$ac_grep_f90_Compaq"; then
  ac_cv_prog_f90_type="Compaq"
  ac_cv_prog_f90_version=$ac_grep_f90_Compaq
elif test -n "$ac_grep_f90_Digital"; then
  ac_cv_prog_f90_type="DEC"
  ac_cv_prog_f90_version=$ac_grep_f90_Digital
elif test -n "$ac_grep_f90_SGI"; then
  ac_cv_prog_f90_type="SGI"
  ac_cv_prog_f90_version=$ac_grep_f90_SGI
elif test -n "$ac_grep_f90_Intel"; then
  ac_cv_prog_f90_type="Intel"
  ac_cv_prog_f90_version=$ac_grep_f90_Intel
else
  ac_cv_prog_f90_type="unknown"
  ac_cv_prog_f90_version="unknown"
fi

rm -f conftest.log
])

dnl If this is not in the cache, it has been set above
AC_CACHE_CHECK([the compiler version],
[ac_cv_prog_f90_version], 
[:])

AC_CACHE_CHECK([whether this compiler permits quadruple precision],
[ac_cv_prog_f90_quadra],
[AC_LANG_SAVE
AC_LANG_FORTRAN90
AC_TRY_COMPILER_NOINCLUDE(dnl
[program conftest
integer, parameter :: d=selected_real_kind(precision(1.)+1, range(1.)+1)
integer, parameter :: q=selected_real_kind(precision(1._d)+1, range(1._d))
real(kind=q) :: x
end program conftest
], ac_cv_prog_f90_quadra, ac_cv_prog_f90_cross)
AC_LANG_RESTORE
])

AC_CACHE_CHECK([whether this compiler permits quadruple complex],
[ac_cv_prog_f90_qcomplex],
[AC_LANG_SAVE
AC_LANG_FORTRAN90
AC_TRY_COMPILER_NOINCLUDE(dnl
[program conftest
integer, parameter :: d=selected_real_kind(precision(1.)+1, range(1.)+1)
integer, parameter :: q=selected_real_kind(precision(1._d)+1, range(1._d))
complex(kind=q) :: x
end program conftest
], ac_cv_prog_f90_qcomplex, ac_cv_prog_f90_qcross)
AC_LANG_RESTORE
])

AC_CACHE_CHECK([whether this compiler accepts PURE functions],
[ac_cv_prog_f90_pure],
[AC_LANG_SAVE
AC_LANG_FORTRAN90
AC_TRY_COMPILER_NOINCLUDE(dnl
[program conftest
contains
  pure function f(x)
    real, intent(in) :: x
    real :: f
    f = x
  end function f
end program conftest
], ac_cv_prog_f90_pure, ac_cv_prog_f90_cross)
AC_LANG_RESTORE
])

])


dnl AC_LANG_FORTRAN90()
AC_DEFUN(AC_LANG_FORTRAN90,
[define([AC_LANG], [FORTRAN90])dnl
ac_ext=f90
ac_compile='${F90-f90} -c $F90FLAGS conftest.$ac_ext 1>&AC_FD_CC'
ac_link='${F90-f90} -o conftest${ac_exeext} $F90FLAGS $LDFLAGS conftest.$ac_ext $LIBS 1>&AC_FD_CC'
cross_compiling=$ac_cv_prog_f90_cross
])

dnl Try to compile, link and execute TEST-PROGRAM.  Set WORKING-VAR to
dnl `yes' if the current compiler works, otherwise set it ti `no'.  Set
dnl CROSS-VAR to `yes' if the compiler and linker produce non-native
dnl executables, otherwise set it to `no'.  Before calling
dnl `AC_TRY_COMPILER()', call `AC_LANG_*' to set-up for the right
dnl language.
dnl
dnl Changes: do not include an #include line
dnl 
dnl AC_TRY_COMPILER_NOINCLUDE(TEST-PROGRAM, WORKING-VAR, CROSS-VAR)
AC_DEFUN(AC_TRY_COMPILER_NOINCLUDE,
[cat > conftest.$ac_ext << EOF
[$1]
EOF
if AC_TRY_EVAL(ac_link) && test -s conftest${ac_exeext}; then
  [$2]=yes
  # If we can't run a trivial program, we are probably using a cross compiler.
  if (./conftest; exit) 2>/dev/null; then
    [$3]=no
  else
    [$3]=yes
  fi
else
  echo "configure: failed program was:" >&AC_FD_CC
  cat conftest.$ac_ext >&AC_FD_CC
  [$2]=no
fi
rm -fr conftest*])



dnl Test for GNU tar:
dnl
dnl AC_PROG_GNU_TAR()
AC_DEFUN(AC_PROG_GNU_TAR,
[AC_CACHE_CHECK([for GNU tar], [ac_cv_prog_GTAR],
[ac_cv_prog_GTAR=$GTAR
test -z "$ac_cv_prog_GTAR" && ac_cv_prog_GTAR=tar
if test -n "`$ac_cv_prog_GTAR --version 2>/dev/null | grep GNU`"; then
  :
elif test -n "`gtar --version 2>/dev/null | grep GNU`"; then
  ac_cv_prog_GTAR=gtar
else
  ac_cv_prog_GTAR=no
fi
])
GTAR=$ac_cv_prog_GTAR
])

dnl Test for GNU make:
dnl
dnl AC_PROG_GNU_MAKE()
AC_DEFUN(AC_PROG_GNU_MAKE,
[AC_CACHE_CHECK([for GNU make], [ac_cv_prog_GMAKE],
[ac_cv_prog_GMAKE="$GMAKE"  # User-supplied command
test -z "$ac_cv_prog_GMAKE" && ac_cv_prog_GMAKE="make"
if test -n "`$ac_cv_prog_GMAKE --version 2>/dev/null | grep GNU`"; then
  :
elif test -n "`gmake --version 2>/dev/null | grep GNU`"; then
  ac_cv_prog_GMAKE=gmake
else
  ac_cv_prog_GMAKE=no
fi
])
GMAKE=$ac_cv_prog_GMAKE
])

dnl Special macros:

dnl Test whether the C compiler as specified in the CompHEP settings 
dnl can compile and link a trivial C program.
dnl 
dnl AC_PROG_COMPHEP_CC_WORKS()
AC_DEFUN(AC_PROG_COMPHEP_CC_WORKS,
[AC_CACHE_CHECK([whether the C compiler selected by CompHEP works], [ac_cv_prog_COMPHEP_CC_works],
[AC_LANG_SAVE
AC_LANG_COMPHEP_CC
AC_TRY_COMPILER_NOINCLUDE([main(){return(0);}], ac_cv_prog_COMPHEP_CC_works, ac_cv_prog_COMPHEP_CC_cross)
AC_LANG_RESTORE
])
if test $ac_cv_prog_COMPHEP_CC_works = no; then
  AC_MSG_ERROR([installation or configuration problem: C compiler cannot create executables.
Check $comphepcom/CC if it is set up correctly])
fi

dnl AC_CACHE_CHECK([whether this compiler is a cross-compiler],
dnl [ac_cv_prog_COMPHEP_CC_cross],
dnl [cross_compiling=$ac_cv_prog_COMPHEP_CC_cross
dnl ])
])

dnl Test whether the F77 compiler as specified in the CompHEP settings 
dnl can compile and link a trivial program.
dnl 
dnl AC_PROG_COMPHEP_F77_WORKS()
AC_DEFUN(AC_PROG_COMPHEP_F77_WORKS,
[AC_CACHE_CHECK([whether the Fortran77 compiler selected by CompHEP works],
[ac_cv_prog_COMPHEP_F77_works],
[AC_LANG_SAVE
AC_LANG_COMPHEP_F77
AC_TRY_COMPILER_NOINCLUDE(dnl
[      program conftest
      end
], ac_cv_prog_COMPHEP_F77_works, ac_cv_prog_COMPHEP_F77_cross)
AC_LANG_RESTORE
])
if test $ac_cv_prog_COMPHEP_F77_works = no; then
  AC_MSG_ERROR([installation or configuration problem: f77 compiler cannot create executables.
Check $comphepcom/F77 if it is set up correctly])
fi
AC_CACHE_CHECK([whether this compiler is a cross-compiler],
[ac_cv_prog_COMPHEP_F77_cross],
[cross_compiling=$ac_cv_prog_COMPHEP_F77_cross
])])



dnl AC_LANG_COMPHEP_CC()
AC_DEFUN(AC_LANG_COMPHEP_CC,
[define([AC_LANG], [COMPHEP_CC])dnl
ac_ext=c
ac_compile='sh $comphepcom/CC conftest.$ac_ext 1>&AC_FD_CC'
ac_link='sh $comphepcom/CC -o conftest${ac_exeext} conftest.$ac_ext $LIBS 1>&AC_FD_CC'
cross_compiling=$ac_cv_prog_COMPHEP_CC_cross
])


dnl AC_LANG_COMPHEP_F77()
AC_DEFUN(AC_LANG_COMPHEP_F77,
[define([AC_LANG], [COMPHEP_F77])dnl
ac_ext=f
ac_compile='sh $comphepcom/F77 conftest.$ac_ext 1>&AC_FD_CC'
ac_link='sh $comphepcom/F77 -o conftest${ac_exeext} conftest.$ac_ext $LIBS 1>&AC_FD_CC'
cross_compiling=$ac_cv_prog_COMPHEP_F77_cross
])



dnl Check for the MPI Fortran 90 compiler. 
dnl Set the output variable `MPIF90' to the name of the compiler found.
dnl If no compiler found, set `MPIF90' to the value of `F90'
dnl 
dnl AC_PROG_MPIF90()
AC_DEFUN(AC_PROG_MPIF90,
[AC_BEFORE([$0], [AC_PROG_CPP])dnl
if test -z "$MPIF90"; then
  AC_PATH_PROGS(MPIF90, mpif95 mpif90)
  test -z "$MPIF90" && AC_MSG_WARN([*** No MPI Fortran 90 compiler found in \$PATH.
Parallel version won't be available.])
fi
if test -n "$MPIF90"; then
  AC_PROG_MPIF90_WORKS
else
  MPIF90=$F90
fi
])

dnl Test whether the MPI Fortran 90 compiler can compile and link a trivial
dnl Fortran program.  Also, test whether the Fortran 90 compiler is a
dnl cross-compiler.
dnl 
dnl AC_PROG_MPIF90_WORKS()
AC_DEFUN(AC_PROG_MPIF90_WORKS,
[AC_MSG_CHECKING([whether the MPI Fortran 90 compiler ($MPIF90 $F90FLAGS $LDFLAGS) works])
AC_LANG_SAVE
AC_LANG_FORTRAN90
AC_TRY_COMPILER_NOINCLUDE(dnl
[program conftest
end program conftest
], ac_cv_prog_mpif90_works, ac_cv_prog_mpif90_cross)
AC_LANG_RESTORE
AC_MSG_RESULT($ac_cv_prog_mpif90_works)

if test $ac_cv_prog_f90_works = no; then
  AC_MSG_ERROR([installation or configuration problem: MPI Fortran 90 compiler cannot create executables.])
fi
])


dnl *** This is probably obsolete:
dnl Check for the f77 runtime library.  This may be needed for linking 
dnl f90 programs with f77 libraries, at least on gcc-based Linux.
dnl We use the LD_LIBRARY_PATH.  Since we are on Linux,
dnl we may try also to find it with 'locate'.
dnl The path part of the result is stored in F77_LIBDIR
dnl If it is not found, return 'no'.  If we are not on Linux, leave this empty.
dnl
dnl AC_CHECK_F77_RUNTIME_LIBRARY()
AC_DEFUN(AC_CHECK_F77_RUNTIME_LIBRARY,
[if test "`uname`" = "Linux"
then
AC_CHECK_PROG(LOCATE,locate,yes)
if test "$LOCATE" = "yes"
then
LD_LIBRARY_PATH=`locate libg2c.a | sed -e 's!/libg2c.a!!' | xargs echo | sed -e 's/ /:/g'`:$LD_LIBRARY_PATH
fi
AC_PATH_PROG(F77_RUNTIME_LIBRARY, libg2c.a, no, $LD_LIBRARY_PATH)
F77_LIBDIR=`echo $F77_RUNTIME_LIBRARY | sed -e s!/libg2c.a!!`
AC_SUBST(F77_LIBDIR)
fi
])


dnl Test whether Metapost works and can dump a .mem file
dnl
dnl AC_PROG_MPOST_WORKS()
AC_DEFUN(AC_PROG_MPOST_WORKS,
[AC_CACHE_CHECK([whether Metapost ($MPOST) works], [ac_cv_prog_mpost_works],
[touch conftest.mp
echo dump | $MPOST --ini conftest >/dev/null 2>&1
if test -f conftest.mem; then
  ac_cv_prog_mpost_works="yes"
else
  ac_cv_prog_mpost_works="no"
fi
rm -f conftest.mp conftest.mem conftest.log

if test $ac_cv_prog_mpost_works = "no"; then
  AC_MSG_ERROR([installation or configuration problem: METAPOST cannot create .mem files.])
fi
dnl AC_MSG_RESULT($ac_cv_prog_mpost_works)
])])

