! WHIZARD 1.21 Feb 26 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module events

  use kinds, only: default !NODEP!
  use limits, only: PROCESS_ID_LEN
  use diagnostics, only: msg_buffer, msg_message, msg_fatal, msg_bug
  use constants
  use lorentz
  use parameters, only: parameter_set !NODEP!
!  use polarization
  use particles
  use density_vectors, only: state_density, create, destroy
  use density_vectors, only: read, write
  use density_vectors, only: read_dimensions_raw, write_dimensions_raw
  use partons, only: parton_state
  use jetset_interface !NODEP!
  use stdhep_interface !NODEP!

  implicit none
  private

  integer, parameter, public :: SCREEN_FMT=0, HEPEVT_FMT=1, RAW_FMT=-1
  integer, parameter, public :: SHORT_FMT=2, LONG_FMT=3
  integer, parameter, public :: JETSET_FMT1=11, JETSET_FMT2=12, JETSET_FMT3=13
  integer, parameter, public :: STDHEP_FMT=20

  type, public :: event
     character(len=PROCESS_ID_LEN) :: process_id
     integer :: process_index
     integer :: count, count_internal, n_in, n_out, n_tot
     logical :: scattering
     real(kind=default) :: sqrts, sqrts_hat, s_hat
     real(kind=default) :: flux_factor, phase_space_factor
     real(kind=default), dimension(2) :: x_parton
     type(particle), dimension(:), pointer :: prt
     real(kind=default) :: mc_function_value, mc_function_ratio
     type(state_density), pointer :: rho_in
     real(kind=default), dimension(:,:,:), pointer :: rho_out
     integer :: n_col_out, n_hel_out, n_flv_out
     integer, dimension(:), pointer :: hel_out, color_flow, anticolor_flow
     type(lorentz_transformation) :: L0
  end type event

  public :: create, destroy
  public :: destroy_rho_in
  public :: read, write
  public :: write_process
  public :: set_in_state
  public :: set_out_state
  public :: get
  public :: merge

  interface create
     module procedure event_create
  end interface
  interface destroy
     module procedure event_destroy
  end interface
  interface write
     module procedure event_write_unit
     module procedure event_write_formatted
  end interface
  interface read
     module procedure event_read_unit
  end interface
  interface write_process
     module procedure write_process_unit
  end interface
  interface set_in_state
     module procedure event_in_state_partons, event_in_state_sqrts
  end interface
  interface set_in_state
     module procedure event_in_state_prt
  end interface
  interface set_out_state
     module procedure event_out_state_codes
     module procedure event_out_state_rho
  end interface
  interface get
     module procedure event_get_particle
  end interface
  interface merge
     module procedure event_apply_decay
  end interface

contains

  subroutine event_create (evt, n_in, n_out, n_col_out, n_hel_out, n_flv_out)
    type(event), intent(out) :: evt
    integer, intent(in) :: n_in, n_out
    integer, intent(in) :: n_col_out, n_hel_out, n_flv_out
    !allocate(evt)
    evt%process_id = " "
    evt%process_index = 0
    evt%count = 0
    evt%count_internal = 0
    evt%scattering = (n_in==2)
    evt%n_in = n_in
    evt%n_out = n_out
    evt%n_tot = n_in + n_out
    if (evt%scattering) then
       allocate(evt%prt(-n_in:n_out))
    else
       allocate(evt%prt(0:n_out))
    end if
    evt%prt = particle_null
    nullify (evt%rho_in)
    evt%n_col_out = n_col_out
    evt%n_hel_out = n_hel_out
    evt%n_flv_out = n_flv_out
    allocate (evt%rho_out(n_col_out, n_hel_out, n_flv_out))
    evt%rho_out = 0
    evt%mc_function_value = 0
    evt%mc_function_ratio = 0
    allocate(evt%hel_out(n_out))
    evt%hel_out = 0
    allocate(evt%color_flow(evt%n_tot), evt%anticolor_flow(evt%n_tot))
    evt%color_flow = 0
    evt%anticolor_flow = 0
  end subroutine event_create

  subroutine event_destroy (evt)
    type(event), intent(inout) :: evt
    deallocate(evt%color_flow, evt%anticolor_flow)
    deallocate(evt%hel_out)
    deallocate(evt%rho_out)
    nullify(evt%rho_in)
    deallocate(evt%prt)
    !deallocate(evt)
  end subroutine event_destroy

  subroutine destroy_rho_in (evt)
    type(event), intent(inout) :: evt
    call destroy (evt%rho_in)
    nullify (evt%rho_in)
  end subroutine destroy_rho_in

  subroutine event_write_unit (u, evt)
    integer, intent(in) :: u
    type(event), intent(in) :: evt
    call write (u, evt, 1._default, SCREEN_FMT)
  end subroutine event_write_unit

  subroutine event_write_formatted (u, evt, weight, fmt, probability, channel)
    integer, intent(in) :: u
    type(event), intent(in) :: evt
    real(kind=default), intent(in) :: weight
    integer, intent(in) :: fmt
    real(kind=default), intent(in), optional :: probability
    integer, intent(in), optional :: channel
    integer :: mode, i, j
    real(kind=default), dimension(0:3) :: p
    mode = fmt
    select case(mode)
    case(SCREEN_FMT)
       write(u,*) '============================================================================'
       write(u,*) 'Event # ', evt%count
       write(u,*) 'process ID = ', evt%process_index, "("//trim(evt%process_id)//")"
       write(u,*) 'count =      ', evt%count_internal
       write(u,*) 'sqrts =      ', evt%sqrts
       write(u,*) 'sqrts_hat =  ', evt%sqrts_hat
       write(u,*) 'flux factor =', evt%flux_factor
       write(u,*) 'phs factor = ', evt%phase_space_factor
       if (evt%scattering) write(u,*) 'x_parton =', evt%x_parton
       write(u,*)
       write(u,*) 'In-state:'
       call write(u,evt%prt(0))
       if (evt%scattering) then
          do i=1,2
             call write(u,evt%prt(-i))
          end do
       end if
       if (associated(evt%rho_in))  call write (u, evt%rho_in)
       write(u,*)
       write(u,*) 'Out-state:'
       do i=1,evt%n_out
          call write(u,evt%prt(i))
       end do
       ! call write(u, evt%rho_out)
       ! write(u,*) 'Squared matrix element:', evt%sqme
       write(u,*) 'MC function value:', evt%mc_function_value
       write(u,*) 'MC function ratio:', evt%mc_function_ratio
       write(u,*) 'Helicity vector:', evt%hel_out
    case(HEPEVT_FMT)
       write(u,11) evt%n_out
       do i=1, evt%n_out
          p = array(particle_momentum(evt%prt(i)))
          write (u,13) 1, particle_code(evt%prt(i)), &
               & (0, j=1,2), (0, j=1,2), evt%hel_out(i)
          write (u,12) p(1:3), p(0), particle_mass(evt%prt(i))
          write (u,12) (0._default, j=1,5)
       end do
    case(SHORT_FMT)
       write(u,11) evt%n_out
       do i=1, evt%n_out
          p = array(particle_momentum(evt%prt(i)))
          write (u,13) particle_code(evt%prt(i)), evt%hel_out(i)
          write (u,12) p(1:3), p(0), particle_mass(evt%prt(i))
       end do
    case(LONG_FMT)
       write (u,11) evt%n_out
       do i=1, evt%n_out
          p = array(particle_momentum(evt%prt(i)))
          write (u,13) particle_code(evt%prt(i)), evt%hel_out(i)
          write (u,12) p(1:3), p(0), particle_mass(evt%prt(i))
       end do
       write (u, 14) evt%mc_function_value, evt%mc_function_ratio
    case(RAW_FMT)
       write(u) evt%process_id, evt%process_index, evt%n_out
       j=0;  if (evt%scattering) j=2
       do i=0,j
          write(u)  particle_code(evt%prt(-i)), &
               & array(particle_momentum(evt%prt(-i))), &
               & particle_mass(evt%prt(-i))
       end do
       do i=1, evt%n_out
          write(u) particle_code(evt%prt(i)), &
               & array(particle_momentum(evt%prt(i))), &
               & particle_mass(evt%prt(i)), &
               & evt%hel_out(i)
          write(u) evt%color_flow(i), evt%anticolor_flow(i)
       end do
       if (present(probability).and.present(channel)) then
          write(u) weight, channel, probability
       else
          write(u) weight
       end if
       if (associated (evt%rho_in)) then
          call write_dimensions_raw (u, evt%rho_in)
          call write (u, evt%rho_in, raw=.true.)
       else
          call msg_bug (" Writing event: in-state density not allocated.")
       end if
       write(u)  evt%flux_factor, evt%phase_space_factor, evt%mc_function_value
    case(STDHEP_FMT)
       call stdhep_write
    case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
       call jetset_write (u,mode-10)
    case default
       call msg_fatal (" Invalid event output format")
    end select
  11 format(1I5)
  12 format(10F17.10)
  13 format(I9,I9,5I5)
  14 format(ES17.10,F17.10)
  end subroutine event_write_formatted

  subroutine event_read_unit &
       & (u, evt, weight, fmt, probability, channel, iostat)
    integer, intent(in) :: u
    type(event), intent(inout) :: evt
    integer, intent(in), optional :: fmt
    real(kind=default), intent(out) :: weight
    real(kind=default), intent(out), optional :: probability
    integer, intent(out), optional :: channel, iostat
    integer :: mode, i, j, code
    real(kind=default) :: mass
    real(kind=default), dimension(0:3) :: p
    integer :: n_col_in, n_hel_in, n_flv_in
    mode = SCREEN_FMT;  if (present(fmt)) mode = fmt
    select case(mode)
    case(RAW_FMT)
       iostat = 0
       read(u, iostat=iostat) evt%process_id, evt%process_index, evt%n_out
       evt%n_tot = evt%n_in + evt%n_out
       j=0;  if (evt%scattering) j=2
       do i=0,j
          read(u, iostat=iostat) code, p, mass
          if (iostat/=0) return
          evt%prt(-i) = particle_new(code, mass, &
               & four_momentum_moving(p(0), three_momentum_moving(p(1:3))))
       end do
       do i=1, evt%n_out
          read(u, iostat=iostat) code, p, mass, evt%hel_out(i)
          read(u, iostat=iostat)  evt%color_flow(i), evt%anticolor_flow(i)
          evt%prt(i) = particle_new(code, mass, &
               & four_momentum_moving(p(0), three_momentum_moving(p(1:3))))
       end do
       if (present(probability).and.present(channel)) then
          read(u) weight, channel, probability
       else
          read(u) weight
       end if
       if (associated (evt%rho_in)) &
            & call msg_fatal (" Reading event: In-state density not properly deallocated.")
       allocate (evt%rho_in)
       call read_dimensions_raw (u, evt%rho_in)
       call create (evt%rho_in)
       call read (u, evt%rho_in, raw=.true.)
       read(u) evt%flux_factor, evt%phase_space_factor, evt%mc_function_value
       evt%mc_function_ratio = 1
    case default
       call msg_fatal (" Can read events in raw format only")
    end select
  11 format(1I5)
  12 format(10F17.10)
  13 format(I9,I9,5I5)
  14 format(ES17.10,F17.10)
  end subroutine event_read_unit

  subroutine write_process_unit (u, process_id, code, n_in, n_out)
    integer, intent(in) :: u
    character(len=*), intent(in) :: process_id
    integer, dimension(:), intent(in) :: code
    integer, intent(in) :: n_in, n_out
    integer :: i, pos
    character(len=len(msg_buffer)) :: tmp_buffer
    call msg_message (" Process " // trim(process_id) // ":", unit=u)
    tmp_buffer = " "
    pos = 1
    do i=1, n_in
       call write_prt (pos, code(i), ibset(0,n_in+n_out-i))
    end do
    msg_buffer (pos+1:) = " ->"
    tmp_buffer (pos+1:) = " ->"
    pos = pos + 3
    do i=1, n_out
       call write_prt (pos, code(n_in+i), ibset(0,i-1))
    end do
    call msg_message (unit=u)
    msg_buffer = tmp_buffer
    call msg_message (unit=u)
  contains
    subroutine write_prt (pos, code, bincode)
      integer, intent(inout) :: pos
      integer, intent(in) :: code, bincode
      character(len=8) :: fmt
      integer :: len
      len = max (len_trim(particle_name(code)), 3)
      write(fmt, '(A5,I1,A2)') "(1x,A", len, ")"
      write(msg_buffer(pos+1:), fmt) trim(particle_name(code))
      write(fmt, '(A5,I1,A2)') "(1x,I", len, ")"
      write(tmp_buffer(pos+1:), fmt) bincode
      pos = pos + len + 1
    end subroutine write_prt
  end subroutine write_process_unit

  subroutine event_in_state_partons (evt, ps, par, ok)
    type(event), intent(inout) :: evt
    type(parton_state), intent(in) :: ps
    type(parameter_set), intent(in) :: par
    logical, intent(out) :: ok
    integer :: root_code, i
    real(kind=default) :: pp, rap, sh, sqsh, lda
    real(kind=default), dimension(2) :: mass
    ok = .true.
    evt%x_parton = ps%x
    sh   = ps%x(1) * ps%x(2) * ps%sqrts**2
    sqsh = sqrt(sh)
    evt%sqrts     = ps%sqrts
    evt%sqrts_hat = sqsh
    evt%s_hat     = sh
    root_code = COMPOSITE
    do i = 1,2
       mass(i) = particle_mass(ps%code(i), par)
    end do
    lda = lambda(evt%s_hat, mass(1)**2, mass(2)**2)
    if (lda <= 0) then
       ok = .false.; return
    end if
    pp = sqrt(lda)
    evt%flux_factor = conv * twopi4 / (2*pp)
    evt%prt(-1) = particle_new( &
         & code = ps%code(1),  mass = mass(1), &
         & p = four_momentum_moving( &
         &       (sqsh + (mass(1)**2-mass(2)**2)/sqsh)/2,  pp/(2*sqsh), 3) )
    evt%prt(-2) = particle_new( &
         & code = ps%code(2),  mass = mass(2), &
         & p = four_momentum_moving( &
         &       (sqsh - (mass(1)**2-mass(2)**2)/sqsh)/2, -pp/(2*sqsh), 3) )
    evt%prt(0) = particle_new( root_code, sqsh )
    rap = ps%rapidity
    if (all(ps%x>0)) then
       rap = rap + log(ps%x(1)/ps%x(2))/2
    else
       ok = .false.; return
    end if
    evt%L0 = boost(sinh(rap),3)
    do i = 0, 2
       evt%prt(-i) = evt%L0 * evt%prt(-i)
    end do
  end subroutine event_in_state_partons

  subroutine event_in_state_sqrts(evt, code, par, sqrts, xi, rapidity, ok)
    type(event), intent(inout) :: evt
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: sqrts
    real(kind=default), dimension(2), intent(in), optional :: xi
    real(kind=default), intent(in), optional :: rapidity
    logical, intent(out), optional :: ok
    integer :: root_code, i, n_in
    real(kind=default) :: pp, rap, sh, sqsh, lda
    real(kind=default), dimension(2) :: mass
    if (present(ok)) ok = .true.
    if (.not.present(xi)) then
       evt%x_parton = 1
       sqsh = sqrts
       sh   = sqrts**2
    else
       evt%x_parton = xi
       sh   = xi(1)*xi(2)*sqrts**2
       sqsh = sqrt(sh)
    end if
    evt%sqrts     = sqrts
    evt%sqrts_hat = sqsh
    evt%s_hat     = sh
    if (evt%scattering) then
       n_in = 2
       root_code = COMPOSITE
       do i = 1,2
          mass(i) = particle_mass(code(i), par)
       end do
       lda = lambda(evt%s_hat, mass(1)**2, mass(2)**2)
       if (lda <= 0) then
          if (present(ok)) then
             ok = .false.; return
          else
             call msg_fatal (" Total energy too low for initial state")
          end if
       end if
       pp = sqrt(lda)
       evt%flux_factor = conv * twopi4 / (2*pp)
       evt%prt(-1) = particle_new( &
            & code = code(1),  mass = mass(1), &
            & p = four_momentum_moving( &
            &       (sqsh + (mass(1)**2-mass(2)**2)/sqsh)/2,  pp/(2*sqsh), 3) )
       evt%prt(-2) = particle_new( &
            & code = code(2),  mass = mass(2), &
            & p = four_momentum_moving( &
            &       (sqsh - (mass(1)**2-mass(2)**2)/sqsh)/2, -pp/(2*sqsh), 3) )
    else
       n_in = 0
       root_code = code(1)
       evt%flux_factor = conv * twopi4 / (2*sqsh)
    end if
    evt%prt(0) = particle_new( root_code, sqsh )
    if (present(rapidity).or.present(xi)) then
       rap = 0
       if (present(rapidity)) rap = rapidity
       if (present(xi)) then
          if (all(xi>0)) then
             rap = rap + log(xi(1)/xi(2))/2
          else
             if (present(ok)) then
                ok = .false.; return
             else
                call msg_fatal (" Can't happen: Non-positive x value")
             end if
          end if
       end if
       evt%L0 = boost(sinh(rap),3)
       do i = 0, n_in
          evt%prt(-i) = evt%L0 * evt%prt(-i)
       end do
    else
       evt%L0 = unit_lorentz_transformation
    end if
  end subroutine event_in_state_sqrts

  subroutine event_in_state_prt(evt, prt, ok)
    type(event), intent(inout) :: evt
    type(particle), dimension(:), intent(in) :: prt
    logical, intent(out), optional :: ok
    integer :: i, n_in
    real(kind=default) :: lda
    type(four_momentum) :: p0
    if (present(ok)) ok = .true.
    if (evt%scattering) then
       n_in = 2
       p0 = particle_momentum(prt(1)) + particle_momentum(prt(2))
       evt%prt(0) = particle_new( code = COMPOSITE,  p = p0 )
       do i=1,2
          evt%prt(-i) = prt(i)
       end do
       lda = lambda(evt%s_hat, &
            &       particle_mass(prt(1))**2, particle_mass(prt(2))**2)
       if (lda<=0) then
          if (present(ok)) then
             ok = .false.; return
          else
             call msg_fatal (" Total energy too low for initial state")
          end if
       end if
       evt%flux_factor = conv * twopi4 / (2 * sqrt(lda))
    else
       n_in = 0
       evt%prt(0) = prt(1)
       evt%flux_factor = conv * twopi4 / (2 * particle_mass(evt%prt(0)))
    end if
    evt%sqrts     = particle_mass(evt%prt(0))
    evt%sqrts_hat = evt%sqrts
    evt%s_hat     = evt%sqrts**2
    evt%L0        = boost( space_part(p0) / evt%sqrts )
    evt%x_parton  = 1
  end subroutine event_in_state_prt

  subroutine event_out_state_codes(evt, code, par)
    type(event), intent(inout) :: evt
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer :: i
    evt%n_out = size (code)
    evt%n_tot = evt%n_in + evt%n_out
    do i=1, evt%n_out
       evt%prt(i) = particle_new( code(i), par)
    end do
  end subroutine event_out_state_codes

  subroutine event_out_state_rho (evt, rho_out, n_col, n_hel, n_flv)
    type(event), intent(inout) :: evt
    real(kind=default), dimension(:,:,:), intent(in) :: rho_out
    integer, intent(in) :: n_col, n_hel, n_flv
    evt%rho_out = 0
    evt%n_col_out = n_col
    evt%n_hel_out = n_hel
    evt%n_flv_out = n_flv
    evt%rho_out(:n_col,:n_hel,:n_flv) = rho_out
  end subroutine event_out_state_rho

  subroutine event_get_particle(evt, prt, code)
    type(event), intent(in) :: evt
    type(particle), intent(out) :: prt
    integer, intent(in) :: code
    integer :: i
    prt = particle_null
    do i=1, evt%n_out
       if (particle_code(evt%prt(i)) == code) then
          prt = evt%prt(i)
          exit
       end if
    end do
  end subroutine event_get_particle
  subroutine event_apply_decay(evt, evt_daughter)
    type(event), intent(inout) :: evt
    type(event), intent(in) :: evt_daughter
    type(particle), dimension(:), pointer :: prt
    integer :: i,n
    n = evt%n_out
    do i=1, n
       if (particle_code(evt%prt(i)) == particle_code(evt_daughter%prt(0))) &
            & then
          n = n + evt_daughter%n_out - 1
          if (evt%scattering) then
             allocate(prt(-2:n))
             prt(-2:-1) = evt%prt(-2:-1)
          else
             allocate(prt(0:n))
          end if
          prt(0:evt%n_out)  = evt%prt(0:)
          prt(i)                  = evt_daughter%prt(1)
          prt(evt%n_out+1:) = evt_daughter%prt(2:)
          deallocate(evt%prt)
          evt%prt => prt
          evt%n_out = n
          exit
       end if
    end do
  end subroutine event_apply_decay


end module events
